from minoshiro.enums import Medium

tables = {
    Medium.ANIME: 'anime',
    Medium.MANGA: 'manga',
    Medium.LN: 'ln',
    Medium.VN: 'vn'
}

convert_medium = {
    'Anime': Medium.ANIME,
    'Manga': Medium.MANGA,
    'LN': Medium.LN
}
